/*****************************************************************************
File:    cpc.c
Purpose: Sample C language control program for Comtrol intelligent
         controllers.  Supports open, close, read, and write to any
         asynchronous line.
Company: Comtrol Corporation
Author:  Craig Harrison
Release: 1.1, 8-23-91 - Original release
         2.0, 9-15-92 - Craig Harrison.  See README file for details.
         2.0.1, 10-21-92 - Craig Harrison.  See README file.
         2.1, 1-28-93 - Craig Harrison.  See README file.
*******************************************************************************
Copyright 1991-1993 Comtrol Corporation. All rights reserved.
Subject to developers license agreement.
*******************************************************************************
                              WARNINGS
All accesses to SCC registers that are paired because of indexing must be
protected by disabling interrupts.  If interrupts are already disabled (as
when in an interrupt service routine) they should not be disabled/enabled
again.  When the ISR exits, the iret instruction will automatically reenable
interrupts.

If modifying this program, do not make calls to any C library functions that
require DOS support.  DOS is not running on the controller.

If using the controller's firmware debugger, interrupt service routines must
not access any near memory.  All pointers must be far pointers, and all
variables must be far.  This is required because the debugger modifies the
DS register while operating, but ISRs will continue to interrupt the debugger.
Therefore ISRs cannot assume that the DS is unchanged while the debugger is
running.
******************************************************************************/


#include <dos.h>
#include "cpc.h"
#include "dpram.h"
#include "firmuser.h"

extern unsigned endcp_seg;             /* segment immediately after ctrl pgm */
char far *heap_p;                      /* ptr to heap */
FIRMUSER_T far *fu_p;                  /* ptr to firmware user area */
MSGQ_T far *comq;                      /* ptr to comq message queue */
MSGQ_T far *sysq;                      /* ptr to sysq message queue */
LINE_ENTRY_T far *line[16];            /* ptrs to line table entries */
char msg_buf[MSG_LEN];                 /* temporary buffer for messages
                                          passed between communications uP
                                          and system uP */
#define TESTMODE 0                     /* set to 1 for line 0 test mode */


/* Function prototypes */
int deq_com_msg(void);
int enq_sys_msg(void);
void ram_init(void);
void vector_init(void);
void set_vector(int,void interrupt far (*)());
void timer_init(int,unsigned,void interrupt far (*)());
void deq_tx_data(LINE_ENTRY_T far *);
void txmode485(LINE_ENTRY_T far *,int);
void interrupt far timer1_isr();
void interrupt far system_isr();
void TBE_isr(LINE_ENTRY_T far *);
void interrupt far line00_TBE();
void interrupt far line01_TBE();
void interrupt far line02_TBE();
void interrupt far line03_TBE();
void interrupt far line04_TBE();
void interrupt far line05_TBE();
void interrupt far line06_TBE();
void interrupt far line07_TBE();
void interrupt far line08_TBE();
void interrupt far line09_TBE();
void interrupt far line10_TBE();
void interrupt far line11_TBE();
void interrupt far line12_TBE();
void interrupt far line13_TBE();
void interrupt far line14_TBE();
void interrupt far line15_TBE();
void RCA_isr(LINE_ENTRY_T far *);
void interrupt far line00_RCA();
void interrupt far line01_RCA();
void interrupt far line02_RCA();
void interrupt far line03_RCA();
void interrupt far line04_RCA();
void interrupt far line05_RCA();
void interrupt far line06_RCA();
void interrupt far line07_RCA();
void interrupt far line08_RCA();
void interrupt far line09_RCA();
void interrupt far line10_RCA();
void interrupt far line11_RCA();
void interrupt far line12_RCA();
void interrupt far line13_RCA();
void interrupt far line14_RCA();
void interrupt far line15_RCA();
void ESC_isr(LINE_ENTRY_T far *);
void interrupt far line00_ESC();
void interrupt far line01_ESC();
void interrupt far line02_ESC();
void interrupt far line03_ESC();
void interrupt far line04_ESC();
void interrupt far line05_ESC();
void interrupt far line06_ESC();
void interrupt far line07_ESC();
void interrupt far line08_ESC();
void interrupt far line09_ESC();
void interrupt far line10_ESC();
void interrupt far line11_ESC();
void interrupt far line12_ESC();
void interrupt far line13_ESC();
void interrupt far line14_ESC();
void interrupt far line15_ESC();
void SRC_isr(LINE_ENTRY_T far *);
void interrupt far line00_SRC();
void interrupt far line01_SRC();
void interrupt far line02_SRC();
void interrupt far line03_SRC();
void interrupt far line04_SRC();
void interrupt far line05_SRC();
void interrupt far line06_SRC();
void interrupt far line07_SRC();
void interrupt far line08_SRC();
void interrupt far line09_SRC();
void interrupt far line10_SRC();
void interrupt far line11_SRC();
void interrupt far line12_SRC();
void interrupt far line13_SRC();
void interrupt far line14_SRC();
void interrupt far line15_SRC();
void null_cmd(void);
void open(void);
void close(void);
void int_sys(void);
void set_testmode(void);


/*****************************************************************************
Function: main
Purpose:  Initialization and main processing loop.
Entry:    Nothing.
Return:   Never returns.
Comments: This is called from module cpcstart.asm, it never returns.
*/
main()
{
   int linenum;                        /* line number */
   LINE_ENTRY_T far *lt_p;             /* ptr to line table entry */
   char far *ram_p;                    /* ptr to some RAM address */

   ram_init();                         /* initialize RAM */
   timer_init(1,CNT200,timer1_isr);    /* start up timer 1, 5ms period */
#if TESTMODE
   set_testmode();                     /* open line 0, this for debug only */
#endif


   /* Main processing loop, handles Tx characters only */
   for(linenum = 0;;linenum = (linenum + 1) % NUMLINES) /* infinite loop */
   {
      lt_p = line[linenum];            /* get ptr to line table entry */
      if(!(lt_p->line_status & LINE_ACTIVE)) /* line active? */
         continue;                     /*    no - move on to next line */
      if(lt_p->line_status & TX_ACTIVE) /* transmit active? */
         continue;                     /*    yes - move on to next line */
      deq_tx_data(lt_p);               /* send char from xmit queue */
   }
}

/*****************************************************************************
Function: ram_init
Purpose:  Initialize RAM. This includes pointer to queues and line tables,
          interrupt vectors, line tables, and the firmware data area
          interaction flag.
Entry:    Nothing.
Return:   Nothing.
*/
void ram_init(void)
{
   int linenum;                        /* line number */
   LINE_ENTRY_T far *lt_p;             /* ptr to line table entry */
   char far *ram_p;                    /* ptr to some RAM address */
   int vector_type;                    /* interrupt vector type */
   static int ioaddr[] =               /* SCC I/O addresses */
   {
#if defined HOSTESSi
      LN00_IO, LN01_IO, LN02_IO, LN03_IO, LN04_IO, LN05_IO, LN06_IO, LN07_IO,
      LN08_IO, LN09_IO, LN10_IO, LN11_IO, LN12_IO, LN13_IO, LN14_IO, LN15_IO
#else if defined HOSTESS186 || defined SMARTH
      LN00_IO, LN01_IO, LN02_IO, LN03_IO, LN04_IO, LN05_IO, LN06_IO, LN07_IO
#endif
   };

   /* Initialize interrupt vector table */
   disable();                    /* disable interrupts while initializing RAM */
   vector_init();

   /* Initialize message queue pointers */
   ram_p = (char far *)COMQ_ADDR;      /* start addr of comq */
   comq = (MSGQ_T far *)ram_p;         /* ptr to comq */
   comq->head = 0;                     /* init head */
   comq->tail = 0;                     /*   and tail */
   ram_p += sizeof(MSGQ_T);
   sysq = (MSGQ_T far *)ram_p;         /* ptr to sysq */
   sysq->head = 0;                     /* init head */
   sysq->tail = 0;                     /*   and tail */
   ram_p += sizeof(MSGQ_T);
   ram_p += 8;    /* start line table on paragarph boundary for easy viewing */

   /* Initialize line table entry pointers */
   for(linenum = 0;linenum < NUMLINES;linenum++,ram_p += sizeof(LINE_ENTRY_T))
      line[linenum] = (LINE_ENTRY_T far *)ram_p;
   ram_p += (16 - linenum) * sizeof(LINE_ENTRY_T); /* if < 16 lines must leave
                                empty space so Tx buffers properly addressed */

   /* Initialize line tables to defaults */
   vector_type = SCCBASE_TYPE;
   for(linenum = 0;linenum < NUMLINES;linenum++)   /* all fields except Rxq */
   {
      lt_p = line[linenum];
      lt_p->io_base = ioaddr[linenum];
      lt_p->line_status = 0;
      lt_p->WR2_ = vector_type;     /* both SCC channels get same base vector*/
      vector_type += (linenum % 2) * 0x10;
      lt_p->WR3_ = 0xc0;
      lt_p->WR4_ = 0x04;
      lt_p->WR5_ = 0x60;
      lt_p->WR12_ = BPS9600;
      lt_p->WR13_ = BPS9600 >> 8;
      lt_p->WR15_ = 0;
      lt_p->Txq_head = 0;
      lt_p->Txq_tail= 0;
      lt_p->Txq_com = ram_p;
      ram_p += TXB_SIZE;
      lt_p->Rxq_head = 0;
      lt_p->Rxq_tail = 0;
   }
   ram_p += (16 - linenum) * TXB_SIZE;  /* if < 16 lines must leave empty space
                                           so Rx buffers properly addressed */
   for(linenum = 0;linenum < NUMLINES;linenum++)   /* do Rxq here */
   {
      line[linenum]->Rxq_com = ram_p;
      ram_p += RXB_SIZE;
   }

   /* Initialize firmware user area and heap pointers */
   enable();                           /* reenable interrupts */
   heap_p = MK_FP(endcp_seg,0);        /* heap ptr */
   fu_p = (FIRMUSER_T far *)FIRMUSER_ADDR;   /* firm user area ptr */
   fu_p->i_flag = 0x55aa;              /* signal host we're alive */
}

/*****************************************************************************
Function: timer_init
Purpose:  Initialize a timer and start it up.  This function initializes
          the timer interrupt vector and timer registers.
Entry:    int tnum: timer number (0, 1, or 2).
          unsigned count: count value to be written to count register.
          void interrupt far (*isr)(): ptr to interrupt service routine.
Return:   Nothing.
*/
void timer_init(int tnum,unsigned count,void interrupt far (*isr)())
{
#if defined HOSTESSi
   static TIMER_T timers[MAX_TIMER-MIN_TIMER+1] = /* information on each timer*/
   {
      TIM0_TYPE, TIM0_CTLWD, TIMER0,
      TIM1_TYPE, TIM1_CTLWD, TIMER1,
      TIM2_TYPE, TIM2_CTLWD, TIMER2
   };
#else if defined HOSTESS186 || defined SMARTH
   static TIMER_T timers[MAX_TIMER-MIN_TIMER+1] = /* information on each timer*/
   {
      TIM0_TYPE, TIM0_CTRL, TIM0_MAX_CNTA,
      TIM1_TYPE, TIM1_CTRL, TIM1_MAX_CNTA
   };
#endif

   if(tnum < MIN_TIMER || tnum > MAX_TIMER)  /* invalid timer number */
      return;

   disable();          /* disable interrupts before modifying vector table */
   set_vector(timers[tnum].vector_type,isr); /* initialize interrupt vector */
#if defined HOSTESSi
   OUTB(TIM_CTL,timers[tnum].control_word);  /* control reg */
   OUTB(timers[tnum].count_reg,count);       /* count least significant byte */
   OUTB(timers[tnum].count_reg,count >> 8);  /* count most significant byte */
#else if defined HOSTESS186 || defined SMARTH
   OUTW(TIM_INT_CTRL,0);                     /* enable timer interrupt mask */
   OUTW(timers[tnum].maxcnta_reg,count);     /* write count value */
   OUTW(timers[tnum].ctrl_reg,TIM_CTLWD);    /* write control word */
#endif
   enable();                           /* reenable interrupts */
}

/*****************************************************************************
Function: vector_init
Purpose:  Initialize interrupt vector table. Each vector requires 4 bytes,
          two for the offset and two for the segment.  The firmware has
          already initialized all unused vectors to point to an "invalid
          interrupt ISR," so it is not necessary to do so again here.
          The SCC vectors are contained in the table below.
          Since the SCC modifies bits 3, 2, and 1 of the base vector type, but
          does not modify bit 0, every second SCC vector is unused.
Entry:    Nothing.
Return:   Nothing.
Comments: Timer interrupt vectors are initialized in timer_init() rather
          than here.
          Interrupts should be disabled before modifying interrupt vectors.
*/
void vector_init(void)
{
   /* Table of SCC interrupt vectors */
   static void interrupt far (*vectors[NUMLINES][4])() =
   {
      {line01_TBE, line01_ESC, line01_RCA, line01_SRC},
      {line00_TBE, line00_ESC, line00_RCA, line00_SRC},
      {line03_TBE, line03_ESC, line03_RCA, line03_SRC},
      {line02_TBE, line02_ESC, line02_RCA, line02_SRC},
      {line05_TBE, line05_ESC, line05_RCA, line05_SRC},
      {line04_TBE, line04_ESC, line04_RCA, line04_SRC},
      {line07_TBE, line07_ESC, line07_RCA, line07_SRC},
#if defined HOSTESS186 || defined SMARTH
      {line06_TBE, line06_ESC, line06_RCA, line06_SRC}
#else if defined HOSTESSi
      {line06_TBE, line06_ESC, line06_RCA, line06_SRC},
      {line09_TBE, line09_ESC, line09_RCA, line09_SRC},
      {line08_TBE, line08_ESC, line08_RCA, line08_SRC},
      {line11_TBE, line11_ESC, line11_RCA, line11_SRC},
      {line10_TBE, line10_ESC, line10_RCA, line10_SRC},
      {line13_TBE, line13_ESC, line13_RCA, line13_SRC},
      {line12_TBE, line12_ESC, line12_RCA, line12_SRC},
      {line15_TBE, line15_ESC, line15_RCA, line15_SRC},
      {line14_TBE, line14_ESC, line14_RCA, line14_SRC}
#endif
   };
   int linenum;                        /* line number */
   int i;
   int vector_type;                    /* interrupt vector type number */

   set_vector(SYS_TYPE,system_isr);    /* initialize system ISR vector */

   /* Initialize the vector table entries for each SCC */
   vector_type = SCCBASE_TYPE;         /* type for 1st SCC */
   for(linenum = 0;linenum < NUMLINES;linenum++)   /* for each line */
   {
      for(i = 0;i < 4;i++)             /* there are 4 vectors for each line */
      {
         set_vector(vector_type,vectors[linenum][i]);
         vector_type += 2;             /* skip unused vector */
      }
   }
}

/*****************************************************************************
Function: set_vector
Purpose:  Initialize an interrupt vector in the interrupt vector table.
Entry:    int vector_type: Interrupt vector type number (0 - 0xff).
          void interrupt far (*isr)(): Ptr to interrupt service routine.
Return:   Nothing.
Comments: The interrupt vector table is located in low memory, from 0:0 to
          0:3ff.  Each vector consists of a two byte offset followed by
          a two byte segment.  Together, this segment:offset pair forms
          a pointer to the interrupt service routine to be executed when
          that interrupt type occurs.  To get the starting address of a
          vector within the vector table the vector type is multiplied
          by 4.
Warning:  Do not call this function unless interrupts are disabled.
*/
void set_vector(int vector_type,void interrupt far (*isr)())
{
   unsigned far *vecttbl;              /* ptr to vector table */

   if(vector_type < 0 || vector_type > 0xff) /* invalid vector type number */
      return;

   vecttbl = (unsigned far *)(vector_type * 4);
   *vecttbl = FP_OFF(isr);
   vecttbl++;
   *vecttbl = FP_SEG(isr);
}

/*****************************************************************************
Function: deq_com_msg
Purpose:  Remove a message from the Communications Processor queue, place
          it in msg_buf[].
Entry:    Nothing.
Return:   1 if message successfully dequeued, else 0.
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
int deq_com_msg(void)
{
   char far *mb_p;                     /* ptr to message byte */
   int mi;                             /* message byte counter */

   if(comq->head == comq->tail)        /* queue empty */
      return(0);                       /* fail */

   mb_p = comq->buf[comq->tail];       /* ptr to 1st byte of message */
   for(mi = 0;mi < MSG_LEN;mi++)       /* put msg in temporary buffer */
      msg_buf[mi] = *mb_p++;
   comq->tail = (comq->tail + 1) % MSGQ_SIZE;   /* bump tail pointer */
   return(1);                          /* success */
}

/*****************************************************************************
Function: enq_sys_msg
Purpose:  Add a message to the System Processor queue, get message from
          msg_buf[].
Entry:    Nothing.
Return:   1 if message successfully enqueued, else 0.
*/
int enq_sys_msg(void)
{
   char far *mb_p;                     /* ptr to message byte */
   int mi;                             /* message byte counter */
   int numopen;                        /* number open message spaces in queue */

   if((numopen = comq->tail - comq->head - 1) < 0) /* get open space in queue*/
      numopen += MSGQ_SIZE;            /* adjust for queue wrap */
   if(numopen <= 0)                    /* queue full */
      return(0);                       /* fail */

   mb_p = comq->buf[comq->tail];       /* ptr to 1st byte of message */
   for(mi = 0;mi < MSG_LEN;mi++)       /* get msg from temporary buffer */
      *mb_p++ = msg_buf[mi];
   comq->head = (comq->head + 1) % MSGQ_SIZE; /* bump head pointer */
   return(1);                          /* success */
}

/*****************************************************************************
Function: deq_tx_data
Purpose:  Remove character from transmit queue, write it to SCC.
          SCC Tx buffer must be empty before calling this function.
Entry:    lt_p - Pointer to line table entry
Return:   Nothing
*/
void deq_tx_data(LINE_ENTRY_T far *lt_p)
{
   int tail;
   if(lt_p->line_status & ALLSENT_PEND)  /* waiting for last Tx bit to clear*/
      return;                       /* don't start next message till it does */
   tail = lt_p->Txq_tail;
   if(lt_p->Txq_head == tail)          /* Tx queue empty */
      return;
   if((lt_p->line_status & (RS485ENABLED|TXMODE)) == RS485ENABLED) /*new msg*/
   {
      disable();
      txmode485(lt_p,1);               /* set transceiver to transmit */
      enable();
   }
   lt_p->line_status |= TX_ACTIVE;     /* indicate transmit active */

   if(lt_p->line_status & RS485ENABLED) /* doing RS-485 */
   {
      disable();  /* ensure ALLSENT_PEND gets set in timely manner */
      OUTB(lt_p->io_base + 2,*(lt_p->Txq_com + tail)); /* write char to SCC */
#if !TESTMODE                          /* in testmode never remove char */
      lt_p->Txq_tail = (tail + 1) % TXB_SIZE;  /* update tail pointer */
#endif
      if(lt_p->Txq_head == lt_p->Txq_tail)  /* Tx queue empty */
      {   /* the next TBE_isr indicates end of message */
         lt_p->line_status |= SEND_FLAG; /* mark it to send flar char next */
      }
      enable();
   }
   else                                /* not doing RS-485 */
   {
      OUTB(lt_p->io_base + 2,*(lt_p->Txq_com + tail)); /* write char to SCC */
#if !TESTMODE                          /* in testmode never remove char */
      lt_p->Txq_tail = (tail + 1) % TXB_SIZE;  /* update tail pointer */
#endif
   }
}

/*****************************************************************************
Function: txmode485
Purpose:  Set the state of the RS-485 transceiver.
Entry:    lt_p - Ptr to line table entry
          state - State of RS-485 transceiver, 0 = receive, 1 = transmit
Return:   Nothing
Comments: The transceiver state is controlled by the RTS line.
Warnings:                ********** WARNING **************
          Interrupts must be disabled before calling this function to
          prevent an interrupt from occuring between the pair of OUTB
          instructions for WR5.  Interrupts are not disabled/enabled within
          this function because it is sometimes called from within an interrupt
          service routine. In that case interrupts are already disabled, and
          we do not want to reenable them here because that would premeture,
          the ISR will enable interrupts during it's iret.
*/
void txmode485(LINE_ENTRY_T far *lt_p,int state)
{
   int scc;                            /* SCC command register address */

   scc = lt_p->io_base;                /* get SCC command reg address */
   OUTB(scc,WR5);                      /* set index to SCC WR5 */
   if(state)                           /* set transceiver to transmit */
   {
#if defined HOSTESSi
      OUTB(scc,lt_p->WR5_ | TX_ENABLE | RTS);
#else if defined HOSTESS186 || defined SMARTH
      OUTB(scc,lt_p->WR5_ | TX_ENABLE);
#endif
      lt_p->line_status |= TXMODE;     /* mark it as in Tx state */
   }
   else                                /* set transceiver to receive */
   {
#if defined HOSTESSi
      OUTB(scc,lt_p->WR5_ | TX_ENABLE);
#else if defined HOSTESS186 || defined SMARTH
      OUTB(scc,lt_p->WR5_ | TX_ENABLE | RTS);
#endif
      lt_p->line_status &= ~TXMODE;     /* mark it as in Rx state */
   }
}

/*****************************************************************************
Function: timer1_isr
Purpose:  Timer 1 interrupt service routine.
          Increments heartbeat counter in firmware user area.
Entry:    Nothing.
Return:   Nothing.
*/
void interrupt far timer1_isr()
{
   fu_p->heartbeat++;                  /* bump heartbeat counter */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: system_isr
Purpose:  Process interrupt from system processor
Entry:    Nothing.
Return:   Nothing.
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void interrupt far system_isr()
{
   /* Set up dispatch table, one function for each Sys uP command */
   static void (*dispatch[NUM_SYSCMD])(void) =
   {
      null_cmd,                        /* 0 */
      open,                            /* 1 */
      close,                           /* 2 */
      int_sys                          /* 3 */
   };

#if defined SMARTH
   OUTB(CLRSYSINT_REG,0);              /* clear Sys uP interrupt */
#endif

   if(!deq_com_msg())                  /* get message from Com uP */
   {
      EOI(INTCTL,EOIVAL);              /* end of interrupt to PIC */
      return;                          /* no message, return */
   }

   if(msg_buf[0] < 0 || msg_buf[0] >= NUM_SYSCMD) /* invalid command */
   {
      EOI(INTCTL,EOIVAL);              /* end of interrupt to PIC */
      return;                          /* no message, return */
   }

   (*dispatch[msg_buf[0]])();          /* execute command function */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: null_cmd (command #0)
Purpose:  Handle unimplemented command
Entry:    Nothing.
Return:   Nothing.
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void null_cmd(void)
{
   ;
}

/*****************************************************************************
Function: open (command #1)
Purpose:  Open a line for asynchronous communications
Entry:     msg_buf+1 = line number
           msg_buf+2 = WR3 parameters   (Rx character size)
           msg_buf+3 = WR4 parameters   (stop bits, parity)
           msg_buf+4 = WR5 parameters   (Tx character size)
           msg_buf+5 = WR12 parameter   (lower byte of BRGTC)
           msg_buf+6 = WR13 parameter   (upper byte of BRGTC)
           msg_buf+7 = RS-485 parameter (0=disable RS485, 1=enable RS485)
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void open(void)
{
   LINE_ENTRY_T far *lt_p;             /* ptr to line table entry */
   unsigned char msgb;                 /* a byte from msg_buf */
   int scc;                            /* SCC command register address */
   int rs485mode;                      /* set if in RS-485 mode */

   lt_p = line[msg_buf[1]];            /* get ptr to line table entry */
   scc = lt_p->io_base;                /* get SCC command reg address */
   lt_p->line_status = 0;              /* start out with line status cleared */
   rs485mode = (int)msg_buf[7];        /* get RS-485 mode */

   /* Configure line table entry using message passed from system */
   lt_p->WR3_ = msg_buf[2] & 0xc0;     /* WR3, isolate Rx char size */
   lt_p->WR4_ = msg_buf[3] & 0x0f;     /* WR4, isolate stop bits/parity*/
   msgb = msg_buf[4] & 0x60;           /* WR5, isolate Tx char size */
   lt_p->WR5_ = (lt_p->WR5_ & (0xff-0x60)) | msgb; /* update WR5 Tx char size*/
   lt_p->WR12_ = msg_buf[5];           /* WR12, BRGT low byte */
   lt_p->WR13_ = msg_buf[6];           /* WR13, BRGT high byte */

   /* Initialize SCC from line table entry */
   OUTB(scc,WR4);                      /* stop bits/parity + clock mode */
   OUTB(scc,lt_p->WR4_ | X16_CLOCK);
   OUTB(scc,WR2);                      /* base interrupt vector type */
   OUTB(scc,lt_p->WR2_);
   OUTB(scc,WR3);                      /* Rx character size */
   OUTB(scc,lt_p->WR3_);
   OUTB(scc,WR5);                      /* Tx char size, modem/break */
   OUTB(scc,lt_p->WR5_ | RTS);         /*     make sure RTS is active */
   OUTB(scc,WR9);                      /* interrupt control */
   OUTB(scc,STATUS_LO+VIS);
   OUTB(scc,WR11);                     /* clock sources */
   OUTB(scc,RX_CLK_EQ_BRG+TX_CLK_EQ_BRG);
   OUTB(scc,WR12);                     /* lower byte of BRGTC */
   OUTB(scc,lt_p->WR12_);
   OUTB(scc,WR13);                     /* upper byte of BRGTC */
   OUTB(scc,lt_p->WR13_);
   OUTB(scc,WR14);                     /* BRG source */
   OUTB(scc,BRG_EQ_SYS_CLK+BRG_ENABLE);
   OUTB(scc,WR3);                      /* enable receive */
   OUTB(scc,lt_p->WR3_ | RX_ENABLE);
   if(rs485mode)                       /* in RS-485 mode */
   {
      lt_p->line_status |= RS485ENABLED;  /* mark line in RS-485 mode */
      txmode485(lt_p,0);               /* enable Tx, set transceiver to Rx */
   }
   else                                /* not in RS-485 mode */
   {
      OUTB(scc,WR5);                      /* enable transmit */
      OUTB(scc,lt_p->WR5_ | TX_ENABLE | RTS);
   }

   /* Enable line */
   lt_p->Txq_head = lt_p->Txq_tail;    /* flush the Tx queue */
   lt_p->Rxq_head = lt_p->Rxq_tail;    /* flush the Rx queue */
   lt_p->line_status |= LINE_ACTIVE;   /* indicate line active */

   /* Enable interrupts */
   OUTB(scc,WR15);                     /* external/status interrupt control */
   OUTB(scc,lt_p->WR15_);
   OUTB(scc,WR0);      /* for insurance reset external status interrupts */
   OUTB(scc,RESET_EXT);
   OUTB(scc,WR1);                      /* interrupt enables */
   OUTB(scc,EXT_INT_ENABLE+TX_INT_ENABLE+PARITY_SPECIAL+RX_INT_ENABLE);
   OUTB(scc,WR9);                      /* master interrupt enable */
   OUTB(scc,MIE+STATUS_LO+VIS);
}

/*****************************************************************************
Function: close (command #2)
Purpose:  Close a serial line
Entry:    msg_buf+1 = line number
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void close(void)
{
   LINE_ENTRY_T far *lt_p;             /* ptr to line table entry */
   int scc;                            /* SCC command register address */

   lt_p = line[msg_buf[1]];            /* get ptr to line table entry */
   scc = lt_p->io_base;                /* get SCC command reg address */

   OUTB(scc,WR1);                      /* disable all interrupts */
   OUTB(scc,0);
   lt_p->line_status = 0;              /* disable all in line_status */
}

/*****************************************************************************
Function: int_sys (command #3)
Purpose:  Generate an interrupt to the system.   This would probably never
          be executed as a command from the system but is included here to
          show how to interrupt the system.
Entry:    Nothing.
Return:   Nothing.
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void int_sys(void)
{
   int i;

   OUTB(C2SINT_REG,C2SINT_LOW);        /* set interrupt line low */
   for(i = 0;i < 100;i++);             /* delay to ensure host sees transition */
   OUTB(C2SINT_REG,C2SINT_HI);         /* set interrupt line high  */
}

/*****************************************************************************
Function: TBE_isr
Purpose:  Common Transmit Buffer Empty Interrupt Service Routine.
          This clears the Tx_active flag in the line table to indicate
          that a character is no longer in the process of being
          transmitted.To keep the time in the ISR short data writes
          to the SCC are handled in the main loop.
Entry:    lt_p - Pointer to line table entry
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void TBE_isr(LINE_ENTRY_T far *lt_p)
{
   int scc;                            /* SCC command register address */

   scc = lt_p->io_base;                /* get SCC command register address */
   if(lt_p->line_status & SEND_FLAG)   /* time to send RS-485 flag char */
   {
      lt_p->line_status &= ~SEND_FLAG;
      lt_p->line_status |= ALLSENT_PEND; /* next TBE_isr set Rx state */
      OUTB(lt_p->io_base + 2,FLAGCHAR); /* write flag char to SCC */
      lt_p->line_status &= ~TX_ACTIVE;  /* indicate no char in SCC Tx buffer */
      OUTB(scc,WR0);                    /* end of interrupt to SCC */
      OUTB(scc,RESET_IUS);
      return;
   }
   if(lt_p->line_status & ALLSENT_PEND) /* RS-485, set line to Rx */
   {
      txmode485(lt_p,0);               /* set transceiver to receive state */
      lt_p->line_status &= ~ALLSENT_PEND; /* no longer waiting for last char */
   }
   lt_p->line_status &= ~TX_ACTIVE;    /* indicate no char in SCC Tx buffer */
   OUTB(scc,WR0);                      /* reset pending Tx interrupt */
   OUTB(scc,RESET_TX_INT);
   OUTB(scc,WR0);                      /* end of interrupt to SCC */
   OUTB(scc,RESET_IUS);
}

/*****************************************************************************
Function: lineXX_TBE
Purpose:  Transmit Buffer Empty Interrupt Service Routine.  There is one of
          these for each line.      Since each line has identical requirements
          on TBE, each of these calls the common function TBE_isr().
Entry:    Nothing.
Return:   Nothing.
*/
void interrupt far line00_TBE()
{
   TBE_isr(line[0]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line01_TBE()
{
   TBE_isr(line[1]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line02_TBE()
{
   TBE_isr(line[2]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line03_TBE()
{
   TBE_isr(line[3]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line04_TBE()
{
   TBE_isr(line[4]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line05_TBE()
{
   TBE_isr(line[5]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line06_TBE()
{
   TBE_isr(line[6]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line07_TBE()
{
   TBE_isr(line[7]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line08_TBE()
{
   TBE_isr(line[8]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line09_TBE()
{
   TBE_isr(line[9]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line10_TBE()
{
   TBE_isr(line[10]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line11_TBE()
{
   TBE_isr(line[11]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line12_TBE()
{
   TBE_isr(line[12]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line13_TBE()
{
   TBE_isr(line[13]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line14_TBE()
{
   TBE_isr(line[14]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line15_TBE()
{
   TBE_isr(line[15]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: ESC_isr
Purpose:  Common External Status Change Interrupt Service Routine.  This
          shows how to get the external status, but doesn't do any
          processing on it.
Entry:    lt_p - Pointer to line table entry
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void ESC_isr(LINE_ENTRY_T far *lt_p)
{
   int scc;                            /* SCC command register address */
   unsigned char status;               /* saves the external status */

   scc = lt_p->io_base;                /* get SCC command register address */
   status = inp(scc);                  /* read the external status */
   status = status;                    /* prevent compiler warning */

   /* Do External Status Change processing here */

   OUTB(scc,WR0);                      /* reset external status interrupts */
   OUTB(scc,RESET_EXT);
   OUTB(scc,WR0);                      /* end of interrupt to SCC */
   OUTB(scc,RESET_IUS);
}

/*****************************************************************************
Function: lineXX_ESC
Purpose:  External Status Change Interrupt Service Routine.  There is one of
          these for each line.   Since each line has identical requirements
          on ESC, each of these calls the common function ESC_isr().
Entry:    Nothing.
Return:   Nothing.
*/
void interrupt far line00_ESC()
{
   ESC_isr(line[0]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line01_ESC()
{
   ESC_isr(line[1]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line02_ESC()
{
   ESC_isr(line[2]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line03_ESC()
{
   ESC_isr(line[3]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line04_ESC()
{
   ESC_isr(line[4]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line05_ESC()
{
   ESC_isr(line[5]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line06_ESC()
{
   ESC_isr(line[6]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line07_ESC()
{
   ESC_isr(line[7]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line08_ESC()
{
   ESC_isr(line[8]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line09_ESC()
{
   ESC_isr(line[9]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line10_ESC()
{
   ESC_isr(line[10]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line11_ESC()
{
   ESC_isr(line[11]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line12_ESC()
{
   ESC_isr(line[12]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line13_ESC()
{
   ESC_isr(line[13]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line14_ESC()
{
   ESC_isr(line[14]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line15_ESC()
{
   ESC_isr(line[15]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: RCA_isr
Purpose:  Common Receive Character Available Interrupt Service Routine.
          Add received character to Rx queue if there is room, else
          throws it away.
Entry:    lt_p - Pointer to line table entry
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void RCA_isr(LINE_ENTRY_T far *lt_p)
{
   int scc;                            /* SCC command register address */
   unsigned char ch;                   /* character read from SCC */
   int head;                           /* Rx queue head pointer */
   int num_full;                       /* number Rx queue locations filled */

   scc = lt_p->io_base;                /* get SCC command register address */
   ch = inp(scc+2);                    /* read character from SCC */
   head = lt_p->Rxq_head;
   if((num_full = head - lt_p->Rxq_tail) < 0)  /* num queue locations full */
   {
      num_full += RXB_SIZE;            /* adjust for queue wrap */
   }
   if(num_full < RXB_SIZE - 1)         /* if Rx queue has empty space */
   {
      *(lt_p->Rxq_com + head) = ch;    /* add received character to queue */
      lt_p->Rxq_head = (head + 1) % RXB_SIZE; /* bump head pointer */
   }

   OUTB(scc,WR0);                      /* end of interrupt to SCC */
   OUTB(scc,RESET_IUS);
}

/*****************************************************************************
Function: lineXX_RCA
Purpose:  Receive Character Available Interrupt Service Routine.  There is one
          of these for each line.  Since each line has identical requirements
          on RCA, each of these calls the common function RCA_isr().
Entry:    Nothing.
Return:   Nothing.
*/
void interrupt far line00_RCA()
{
   RCA_isr(line[0]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line01_RCA()
{
   RCA_isr(line[1]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line02_RCA()
{
   RCA_isr(line[2]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line03_RCA()
{
   RCA_isr(line[3]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line04_RCA()
{
   RCA_isr(line[4]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line05_RCA()
{
   RCA_isr(line[5]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line06_RCA()
{
   RCA_isr(line[6]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line07_RCA()
{
   RCA_isr(line[7]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line08_RCA()
{
   RCA_isr(line[8]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line09_RCA()
{
   RCA_isr(line[9]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line10_RCA()
{
   RCA_isr(line[10]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line11_RCA()
{
   RCA_isr(line[11]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line12_RCA()
{
   RCA_isr(line[12]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line13_RCA()
{
   RCA_isr(line[13]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line14_RCA()
{
   RCA_isr(line[14]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line15_RCA()
{
   RCA_isr(line[15]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: SRC_isr
Purpose:  Common Special Receive condition Interrupt Service Routine.
          This shows how to get the special receive condition status,
          but doesn't do any processing on it.
Entry:    lt_p - Pointer to line table entry
Return:   Nothing
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.

*/
void SRC_isr(LINE_ENTRY_T far *lt_p)
{
   int scc;                            /* SCC command register address */
   unsigned status;                    /* saves the SRC status */

   scc = lt_p->io_base;                /* get SCC command register address */
   OUTB(scc,RR1);                      /* read the SRC status */
   status = INB(scc);                  /* read the external status */
   status = status;                    /* prevent compiler warning */

   /* Do Special Receive Condition processing here */

   OUTB(scc,WR0);                      /* for insurance */
   OUTB(scc,ERROR_RESET);              /*   issue error reset command */
   OUTB(scc,WR0);                      /* end of interrupt to SCC */
   OUTB(scc,RESET_IUS);
}

/*****************************************************************************
Function: lineXX_SRC
Purpose:  Special Receive Condition Interrupt Service Routine.   There is one
          of these for each line.  Since each line has identical requirements
          on SRC, each of these calls the common function SRC_isr().
Entry:    Nothing.
Return:   Nothing.
Warning:  Interrupts are disabled within ISRs, do not reenable interrupts
          within this function.  See warning at top of this source file.
*/
void interrupt far line00_SRC()
{
   SRC_isr(line[0]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line01_SRC()
{
   SRC_isr(line[1]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line02_SRC()
{
   SRC_isr(line[2]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line03_SRC()
{
   SRC_isr(line[3]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line04_SRC()
{
   SRC_isr(line[4]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line05_SRC()
{
   SRC_isr(line[5]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line06_SRC()
{
   SRC_isr(line[6]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line07_SRC()
{
   SRC_isr(line[7]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line08_SRC()
{
   SRC_isr(line[8]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line09_SRC()
{
   SRC_isr(line[9]);                   /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line10_SRC()
{
   SRC_isr(line[10]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line11_SRC()
{
   SRC_isr(line[11]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line12_SRC()
{
   SRC_isr(line[12]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line13_SRC()
{
   SRC_isr(line[13]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line14_SRC()
{
   SRC_isr(line[14]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

void interrupt far line15_SRC()
{
   SRC_isr(line[15]);                  /* do common processing */
   EOI(INTCTL,EOIVAL);                 /* end of interrupt to PIC */
}

/*****************************************************************************
Function: set_testmode
Purpose:  Open line 0 for debugging purposes.  This will cause line 0
          to immediately start transmitting a single character continuously.
Entry:    Nothing.
Return:   Nothing.
*/
void set_testmode(void)
{
   LINE_ENTRY_T far *lt_p;             /* ptr to line table entry */

   lt_p = line[0];
   msg_buf[0] = 1;
   msg_buf[1] = 0;
   msg_buf[2] = 0xc0;
   msg_buf[3] = 0x44;
   msg_buf[4] = 0x60;
   msg_buf[5] = 0x0e;
   msg_buf[6] = 0;
   open();
   lt_p->Txq_head = 1;                 /* put char to xmit in Tx buffer */
   lt_p->Txq_tail = 0;
   *lt_p->Txq_com = 0x30;
}
