/******************************************************************************
File:    hiterm.c
Purpose: Source code for a DOS sample application program that runs on the
         system processor.  This program demonstrates how the system processor
         interacts with the control program running on the controller, using
         the open(), close(), read(), and write() functions contained in
         HILIB.C.  HITERM.C and HILIB.C are compiled to form the DOS application
         program HITERM.EXE.  HITERM is a simple terminal emulation program, it
         allows the user to send and receive characters through any serial line
         on the controller.
Company: Comtrol Corporation
Author:  Lori Slater, Craig Harrison
Release: 1.00, 8-23-91 - Original release
         2.0, 9-18-92 - Craig Harrison
            Fixed read loop, was bumping cnt by 2, should be 1.
            Made F9 key send a 40 character string.
         2.0.1, 10-21-92 - Craig Harrison
            Added "F9 to send string" message.
         2.1, 1-28-93 - Craig Harrison
            Changed F9 send string message from decimal digits to "Comtrol ..."
*******************************************************************************
 Copyright 1991-1993 Comtrol Corporation. All rights reserved. Subject to
 developers license agreement.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>

/* Function prototypes */
int hiopen(int linenum);
int hiclose(int linenum);
int hiread(char *buffer,int cnt,int linenum);
int hiwrite(char *buffer,int cnt,int linenum);
void dpram_init(void);

main()
{
   int i;
   int pnum;
   int cnt;
   char buf[80];

   fprintf(stderr,"Serial Line Number (0-15): ");
   gets(buf);                                      /* get serial line */
   sscanf(buf,"%d",&pnum);
   system("cls");                                  /* clear screen */
   fprintf(stderr,"Serial Line Number %d\t\t\tHit F10 to Quit, F9 to send string\n",
   pnum);

   dpram_init();
   if(!hiopen(pnum))
   {
      fprintf(stderr,"Can't open line %d\n",pnum);
      exit(0);
   }

   while(1)                            /* infinite loop */
   {
      /* attempt to read char from device and write to screen */
      if((cnt = hiread(buf,80,pnum)) > 0)
      {
         for(i=0;i<cnt;i++)
         {
            if(buf[i] == 015)          /* CR = CRLF */
               fprintf(stderr,"\n");
            else if(buf[i] != 012)     /* not LF */
               putc(buf[i],stderr);
         }
      }

      /* attempt to read char from keyboard and write to device */
      if((bdos(11,0,0)&0xff) == 0xFF)  /* if char waiting */
      {
         buf[0] = bdos(8,0,0) & 0xff;  /* read keybd char */
         if ((buf[0] == NULL) &&       /* 2 char key */
         ((bdos(11,0,0)&0xff) == 0xff))
         {
            buf[1] = bdos(8,0,0) & 0xff;  /* next */
            if(buf[1] == 0x44)      /* F10 = quit */
               break;
            else if(buf[1] == 0x43) /* F9 = send string */
            {
               strcpy(buf,"Comtrol Hostess & Hostess i Controllers!");
               hiwrite(buf,40,pnum);
            }
            else
               hiwrite(buf,2,pnum);
         }
         else if(buf[0] == 015)         /* CR = CRLF */
         {
            buf[1] = 012;
            hiwrite(buf,2,pnum);
         }
         else if(buf[0] != 012)         /* not LF */
            hiwrite(buf,1,pnum);
      }
   }
   hiclose(pnum);
   return(0);
}
